#define _GNU_SOURCE
#define BUF_NUM 2

#include "gettext.h"
#include "print_error.h"
#include "struct_CPInfo.h"
#include "struct_OPArg.h"
#include "verify_md5sum.h"
#include "write_log_hash_check.h"

#include "md5.h"

#include <stdbool.h>
#include <stdlib.h>

#include <fcntl.h>
#include <malloc.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

static void *buf[BUF_NUM];
static int fd;
static int buffer_size_local;
static long psize;
static md5_state_t state;
static int read_size[BUF_NUM];
static off_t total;
static int r_number;
static int a_number;
static _Bool read_err;
static _Bool read_finish;

/* 関数プロトタイプ */
void verify_hash(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd);
static void o_direct(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd);
static void * read_func(void *dummy);
static void * append_func(void *dummy);

/*******************************************************************************
*******************************************************************************/
void verify_hash(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd)
{
	fd = 0;
	buffer_size_local = oparg->buffer_size / 2;
	read_size[0] = 0;
	read_size[1] = 0;
	total = 0;
	read_err = false;
	read_finish = false;

	if(psize == 0)
	{
		psize = sysconf(_SC_PAGESIZE);
	}

	_Bool is_open;

	errno = 0;
	if((fd = open(cpinfo->dst, O_NOATIME | O_NOFOLLOW | O_DIRECT | O_RDONLY)) == -1)
	{
		switch(errno)
		{
		case EPERM:
			if((fd = open(cpinfo->dst, O_NOFOLLOW | O_DIRECT | O_RDONLY)) == -1)
			{
				is_open = false;
			}
			else
			{
				is_open = true;
			}
			break;

		default:
			is_open = false;
			break;
		}
	}
	else
	{
		is_open = true;
	}

	if(is_open == true)
	{
		o_direct(oparg, sdir, cpinfo, cpdd);
	}
	else
	{
		verify_md5sum(oparg, sdir, cpinfo, cpdd);
	}

	if((is_open == true) && (close(fd) == -1))
	{
		print_error("close", __FILE__, __LINE__, cpdd);
		fprintf(stderr, _("error : %s\n"), cpinfo->dst);
	}
/*
	free(buf[0]);
	free(buf[1]);
*/
}

/*******************************************************************************
*******************************************************************************/
static void o_direct(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd)
{
	static _Bool is_get_memory = false;

	if((buf[0] == NULL) && (buf[1] == NULL))
	{
		if((posix_memalign((void **)&buf[0], psize, buffer_size_local) != 0) || (posix_memalign((void **)&buf[1], psize, buffer_size_local) != 0))
		{
			print_error("posix_memalign", __FILE__, __LINE__, cpdd);
			fprintf(stderr, _("verify error : %s\n"), cpinfo->dst);
			write_log_hash_check(cpinfo, oparg, cpinfo->src_md5, "NULL", sdir, cpdd, false);
			cpinfo->dst_para = INITIAL;

			if(buf[0] != NULL)
			{
				free(buf[0]);
				buf[0] = NULL;
			}

			if(buf[1] != NULL)
			{
				free(buf[1]);
				buf[1] = NULL;
			}
		}
		else
		{
			is_get_memory = true;
		}
	}

	if(is_get_memory == true)
	{
		md5_init(&state);

		r_number = 0;
		a_number = 1;
		pthread_t t1;
		pthread_t t2;

		pthread_create(&t1, NULL, read_func, (void *)&r_number);
		pthread_join(t1, NULL);

		int swap;

		for(;;)
		{
			swap = r_number;
			r_number = a_number;
			a_number = swap;

			pthread_create(&t1, NULL, read_func, (void *)&r_number);
			pthread_create(&t2, NULL, append_func, (void *)&a_number);

			pthread_join(t1, NULL);
			pthread_join(t2, NULL);

			if((read_finish == true) || (read_err == true))
			{
				break;
			}
		}

		md5_byte_t digest[16];
		char hex_output[(16 * 2) + 1];
		hex_output[0] = '\0';

		md5_finish(&state, digest);

		for(int di = 0; di < 16; ++di)
		{
			sprintf(hex_output + di * 2, "%02x", digest[di]);
		}

		_Bool is_equal = true;

		if(strcmp(cpinfo->src_md5, hex_output) == 0)
		{
			cpinfo->hash_check = true;

			if(oparg->V == VERBOS)
			{
				if(oparg->M == NOT)
				{
					printf(_("%s : %s\n"), cpinfo->src_md5, cpinfo->src);
				}

				printf(_("%s : %s\n"), hex_output, cpinfo->dst);
				fflush(stdout);
			}

			if(oparg->L == LOG)
			{
				write_log_hash_check(cpinfo, oparg, cpinfo->src_md5, hex_output, sdir, cpdd, true);
			}
		}
		else if(cpinfo->src_size > total)
		{
			is_equal = false;
			fprintf(stderr, _("read error : %s\n"), cpinfo->dst);
		}
		else
		{
			is_equal = false;
			fprintf(stderr, _("hash error : %s\n"), cpinfo->dst);
		}

		if(is_equal == false)
		{
			print_error("NULL", __FILE__, __LINE__, cpdd);
			fprintf(stderr, _("verify error : %s\n"), cpinfo->dst);
			write_log_hash_check(cpinfo, oparg, cpinfo->src_md5, hex_output, sdir, cpdd, false);
			cpinfo->dst_para = DELETE;

			errno = 0;
			if(unlink(cpinfo->dst) == -1)
			{
				if(errno != ENOENT)
				{
					print_error("unlink", __FILE__, __LINE__, cpdd);
					fprintf(stderr, _("error : %s\n"), cpinfo->dst);
				}
			}
			else
			{
				fprintf(stderr, _("remove : %s\n"), cpinfo->dst);
			}
		}
	}
}

/*******************************************************************************
*******************************************************************************/
static void * read_func(void *dummy)
{
	int i = r_number;

	/* 
	 * O_DIRECTを使用するとアライメントの調整などが面倒なので、
	 * full_read や safe_read のようなエラー処理は行わない。
	*/
	read_size[i] = read(fd, buf[i], buffer_size_local);

	if(read_size[i] == 0)
	{
		read_finish = true;
	}
	else if(read_size[i] < 0)
	{
		read_err = true;
	}
	else
	{
		total += read_size[i];
	}

	dummy++;	/* コンパイル時のwarningを出さないようにするため */
	return NULL;
}

/*******************************************************************************
*******************************************************************************/
static void * append_func(void *dummy)
{
	int i = a_number;

	md5_append(&state, buf[i], read_size[i]);

	dummy++;	/* コンパイル時のwarningを出さないようにするため */
	return NULL;
}
